"""Account Management plugin for VSS CLI (vss-cli)."""
import click
from vss_cli import const
from vss_cli.cli import pass_context
from vss_cli.config import Configuration
from vss_cli.helper import format_output


@click.group(
    'account',
    short_help='Manage your VSS account'
)
@pass_context
def cli(ctx: Configuration):
    """Manage your VSS account."""
    with ctx.spinner(disable=ctx.debug):
        ctx.load_config()


@cli.group('get',
           short_help='get account attribute')
@pass_context
def account_get(ctx: Configuration):
    """Obtain an account attribute"""
    pass


@account_get.group('digest')
@pass_context
def account_get_digest(ctx):
    """Get digest status"""
    pass


@account_get_digest.command('message')
@pass_context
def account_get_digest_message(ctx: Configuration):
    """Get message digest status"""
    with ctx.spinner(disable=ctx.debug):
        _obj = ctx.get_user_digest_settings()
    obj = {'message': _obj.get('message')}
    columns = ctx.columns or const.COLUMNS_MESSAGE_DIGEST
    click.echo(
        format_output(
            ctx,
            [obj],
            columns=columns,
            single=True
        )
    )


@account_get.command('groups')
@pass_context
def account_get_groups(ctx: Configuration):
    """User group membership"""
    with ctx.spinner(disable=ctx.debug):
        obj = dict(groups=ctx.get_user_groups())
    columns = ctx.columns or const.COLUMNS_GROUPS
    click.echo(
        format_output(
            ctx,
            [obj],
            columns=columns,
            single=True
        )
    )


@account_get.command('group')
@click.argument(
    'group_name', type=click.STRING, required=True)
@pass_context
def account_get_group(ctx: Configuration, group_name):
    """Get given group info or members.
    User must be part of the group."""
    with ctx.spinner(disable=ctx.debug):
        obj = ctx.get_user_group(group_name, True)
    columns = ctx.columns or const.COLUMNS_GROUP
    click.echo(
        format_output(
            ctx,
            [obj],
            columns=columns,
            single=True
        )
    )


@account_get.command('access-role')
@pass_context
def account_get_access_role(ctx: Configuration):
    """Access role and entitlements"""
    with ctx.spinner(disable=ctx.debug):
        obj = ctx.get_user_roles()
    columns = ctx.columns or const.COLUMNS_ROLE
    click.echo(
        format_output(
            ctx,
            [obj['access']],
            columns=columns,
            single=True
        )
    )


@account_get.command('request-role')
@pass_context
def account_get_request_role(ctx: Configuration):
    """Request role and entitlements"""
    with ctx.spinner(disable=ctx.debug):
        obj = ctx.get_user_roles()
    columns = ctx.columns or const.COLUMNS_ROLE
    click.echo(
        format_output(
            ctx,
            [obj['request']],
            columns=columns,
            single=True
        )
    )


@account_get.command('personal')
@pass_context
def account_get_personal(ctx: Configuration):
    """User information"""
    with ctx.spinner(disable=ctx.debug):
        obj = ctx.get_user_personal()
    obj.update(ctx.get_user_ldap())
    columns = ctx.columns or const.COLUMNS_USER_PERSONAL
    click.echo(
        format_output(
            ctx,
            [obj],
            columns=columns,
            single=True
        )
    )


@account_get.command('status')
@pass_context
def account_get_pstatus(ctx: Configuration):
    """Account status"""
    with ctx.spinner(disable=ctx.debug):
        obj = ctx.get_user_status()
    columns = ctx.columns or const.COLUMNS_USER_STATUS
    click.echo(
        format_output(
            ctx,
            [obj],
            columns=columns,
            single=True
        )
    )


@account_get.group('notification')
@pass_context
def account_get_notification(ctx: Configuration):
    """Notification settings"""
    pass


@account_get_notification.command('request')
@pass_context
def account_get_notification_request(ctx: Configuration):
    """Get notification format"""
    with ctx.spinner(disable=ctx.debug):
        obj = ctx.get_user_request_notification_settings()
    columns = ctx.columns or const.COLUMNS_NOT_REQUEST
    click.echo(
        format_output(
            ctx,
            [obj],
            columns=columns,
            single=True
        )
    )


@account_get_notification.command('format')
@pass_context
def account_get_notification_format(ctx: Configuration):
    """Get notification format"""
    with ctx.spinner(disable=ctx.debug):
        obj = ctx.get_user_notification_format()
    click.echo(
        format_output(
            ctx,
            [obj],
            columns=ctx.columns,
            single=True
        )
    )


@account_get_notification.command('method')
@pass_context
def account_get_notification_method(ctx: Configuration):
    """Get notification format"""
    with ctx.spinner(disable=ctx.debug):
        obj = ctx.get_user_notification_method()
    click.echo(
        format_output(
            ctx,
            [obj],
            columns=ctx.columns,
            single=True
        )
    )


@cli.group(
    'set',
    short_help='set account attribute'
)
@pass_context
def account_set(ctx: Configuration):
    """Set account attribute"""
    pass


@account_set.group(
    'digest',
    short_help='set account weekly digest configuration'
)
@pass_context
def account_set_digest(ctx: Configuration):
    """update weekly digest configuration"""
    pass


@account_set_digest.command(
    'message'
)
@click.argument(
    'state',
    type=click.Choice(['in', 'out']),
    required=True
)
@pass_context
def account_set_digest_message(
        ctx: Configuration, state
):
    """Opt-in or opt-out of weekly message digest"""
    with ctx.spinner(disable=ctx.debug):
        if state == 'in':
            ctx.enable_user_message_digest()
        else:
            ctx.disable_user_message_digest()
    _obj = ctx.get_user_digest_settings()
    obj = {'message': _obj.get('message')}
    columns = ctx.columns or const.COLUMNS_MESSAGE_DIGEST
    click.echo(
        format_output(
            ctx,
            [obj],
            columns=columns,
            single=True
        )
    )


@account_set.group(
    'notification',
    short_help='set account notification settings'
)
@pass_context
def account_notification_set(ctx: Configuration):
    """Set account notification settings"""
    pass


@account_notification_set.command(
    'request',
)
@click.argument(
    'notification_type',
    type=click.Choice(
        ['all', 'none',
         'error', 'completion',
         'submission']
    ),
    nargs=-1,
    required=True
)
@pass_context
def account_notification_set_request(
        ctx: Configuration,
        notification_type
):
    """Customize request notification settings"""
    lookup = {
        'all': ctx.enable_user_request_all_notification,
        'none': ctx.disable_user_request_all_notification,
        'error': ctx.enable_user_request_error_notification,
        'submission': ctx.enable_user_request_submission_notification,
        'completion': ctx.enable_user_request_completion_notification

    }
    for n_type in notification_type:
        try:
            f = lookup[n_type]
            f()
            if n_type in ['all', 'none']:
                status = 'enabled' if n_type == 'all' else 'disabled'
                ctx.secho(
                    f'Notifications triggered by requests '
                    f'have been {status}.', fg='green'
                )
            elif n_type in ['error', 'submission', 'completion']:
                ctx.secho(
                    f'Notifications triggered by request {n_type} '
                    f'have been enabled.',
                    fg='green'
                )
        except KeyError:
            pass
    with ctx.spinner(disable=ctx.debug):
        obj = ctx.get_user_request_notification_settings()
    columns = ctx.columns or const.COLUMNS_NOT_REQUEST
    click.echo(
        format_output(
            ctx,
            [obj],
            columns=columns,
            single=True
        )
    )


@account_notification_set.command(
    'format'
)
@click.argument(
    'fmt',
    type=click.Choice(['html', 'text']),
    required=True)
@pass_context
def account_notification_set_format(
        ctx: Configuration,
        fmt
):
    """Update notification format where FMT can be html or text"""
    with ctx.spinner(disable=ctx.debug):
        obj = ctx.update_user_notification_format(fmt)
    click.echo(
        format_output(
            ctx,
            [obj],
            columns=ctx.columns,
            single=True
        )
    )


@account_notification_set.command(
    'method'
)
@click.argument(
    'method',
    type=click.Choice(['mail', 'message']),
    required=True
)
@pass_context
def account_notification_set_method(
        ctx: Configuration,
        method
):
    """Update notification method where METHOD can be mail or message"""
    with ctx.spinner(disable=ctx.debug):
        obj = ctx.update_user_notification_method(method)
    click.echo(
        format_output(
            ctx,
            [obj],
            columns=ctx.columns,
            single=True
        )
    )



import logging

import click

from vss_cli import const, rel_opts as so
from vss_cli.cli import pass_context
from vss_cli.config import Configuration
from vss_cli.helper import format_output
from vss_cli.plugins.compute import cli

_LOGGING = logging.getLogger(__name__)


@cli.group('template', short_help='List virtual machine templates')
@pass_context
def compute_template(ctx):
    """List virtual machine templates"""
    pass


@compute_template.command('ls', short_help='List virtual machine templates')
@so.filter_opt
@so.all_opt
@so.page_opt
@so.sort_opt
@so.count_opt
@pass_context
def compute_template_ls(
    ctx: Configuration, filter_by, show_all, sort, page, count
):
    """List virtual machine templates.

        Filter and sort list by any attribute. For example:

        vss-cli compute template ls -f name=like,%vm-name% -f version=like,%13

        Simple name filtering:

        vss-cli compute template ls -f name=%vm-name% -s name=desc

    """
    params = dict(expand=1, sort='name,asc')
    if all(filter_by):
        params['filter'] = ';'.join(filter_by)
    if all(sort):
        params['sort'] = ';'.join(sort)
    # get templates
    with ctx.spinner(disable=ctx.debug):
        obj = ctx.get_templates(show_all=show_all, per_page=count, **params)
    # including additional attributes?
    columns = ctx.columns or const.COLUMNS_VM_TEMPLATE
    # format output
    output = format_output(ctx, obj, columns=columns)
    # page
    if page:
        click.echo_via_pager(output)
    else:
        click.echo(output)



"""Compute plugin for VSS CLI (vss-cli)."""
import click
from vss_cli.cli import pass_context
from vss_cli.config import Configuration
from pkg_resources import iter_entry_points
from click_plugins import with_plugins


@with_plugins(
    iter_entry_points('vss_cli.contrib.compute')
)
@click.group(
    'compute',
    short_help='Manage VMs, networks, folders, etc.'
)
@pass_context
def cli(ctx: Configuration):
    """Compute related resources such as virtual machines, networks
       supported operating systems, logical folders, OVA/OVF images,
       floppy images, ISO images and more."""
    with ctx.spinner(disable=ctx.debug):
        ctx.load_config()


from vss_cli.plugins.compute_plugins import (
    domain, inventory, net, os as compute_os,
    image, iso, floppy, folder, template, vm
)  # pylint: disable=unused-import



"""Constants used by VSS CLI (vss-cli)."""
import os

import pkg_resources

PACKAGE_NAME = "vss_cli"

__version__ = "0.6.2-dev5"

REQUIRED_PYTHON_VER = (3, 6, 4)

DEFAULT_TIMEOUT = 30
DEFAULT_ENDPOINT = "https://cloud-api.eis.utoronto.ca"
DEFAULT_ENDPOINT_NAME = "cloud-api"
DEFAULT_WEBDAV_SERVER = "https://vskey-stor.eis.utoronto.ca"
_LEGACY_CONFIG = ("~", ".vss-cli", "config.json")
_DEFAULT_CONFIG = ("~", ".vss-cli", "config.yaml")
_DEFAULT_HISTORY = ("~", ".vss-cli", "history")

COLUMNS_WIDTH_DEFAULT = -1
COLUMNS_WIDTH_STR = "\u2026"

LEGACY_CONFIG = os.path.expanduser(os.path.join(*_LEGACY_CONFIG))
DEFAULT_CONFIG = os.path.expanduser(os.path.join(*_DEFAULT_CONFIG))
DEFAULT_HISTORY = os.path.expanduser(os.path.join(*_DEFAULT_HISTORY))
DEFAULT_DATA_PATH = pkg_resources.resource_filename(PACKAGE_NAME, "data")
DEFAULT_CONFIG_TMPL = os.path.join(DEFAULT_DATA_PATH, "config.yaml")
DEFAULT_CHECK_UPDATES = True
DEFAULT_CHECK_MESSAGES = True

DEFAULT_TABLE_FORMAT = "simple"
DEFAULT_DATA_OUTPUT = "table"
DEFAULT_RAW_OUTPUT = "json"
DEFAULT_OUTPUT = "auto"
DEFAULT_VERBOSE = False
DEFAULT_DEBUG = False

DEFAULT_SETTINGS = {
    "endpoint": DEFAULT_ENDPOINT,
    "output": DEFAULT_OUTPUT,
    "table_format": DEFAULT_TABLE_FORMAT,
    "check_for_messages": DEFAULT_CHECK_MESSAGES,
    "check_for_updates": DEFAULT_CHECK_UPDATES,
    "timeout": DEFAULT_TIMEOUT,
    "verbose": DEFAULT_VERBOSE,
    "debug": DEFAULT_DEBUG,
}

DEFAULT_DATETIME_FMT = "%Y-%m-%d %H:%M"
SUPPORTED_DATETIME_FORMATS = ["%Y-%m-%dT%H:%M:%S", "%Y-%m-%d %H:%M"]
GENERAL_SETTINGS = {
    "check_for_messages": bool,
    "check_for_updates": bool,
    "debug": bool,
    "verbose": bool,
    "default_endpoint_name": str,
    "output": str,
    "table_format": str,
    "timeout": int,
    "columns_width": int,
}

DEFAULT_HOST_REGEX = (
    "^[a-z][a-z0-9+\\-.]*://([a-z0-9\\"
    "-._~%!$&'()*+,;=]+@)?([a-z0-9\\-."
    "_~%]+|\\[[a-z0-9\\-._~%!$&'()*+,;"
    "=:]+\\])"
)

DEFAULT_NIC_DEL_MSG = (
    "Network adapter:\t{unit} ({type})\n"
    "Mac address:\t\t{mac_address}\n"
    "Network:\t\t{network[name]} ({network[moref]})\n"
    "Connected:\t\t{connected}\n"
)

DEFAULT_STATE_MSG = (
    "Host Name:\t{hostname} ({os[full_name]})\n"
    "IP Address:\t{ip_addresses}\n"
    "Are you sure you want to change the state from "
    '"{guest_state} to {state}" '
    "of the above VM?"
)

DEFAULT_VM_DEL_MSG = (
    "Name:\t\t{name[name]}\n"
    "Folder:\t\t{folder_info[path]}\n"
    "Host Name:\t{hostname} "
    "({os[full_name]})\n"
    "IP Address:\t{ip_addresses}\n"
    "Are you sure you want to delete "
    "the above VM?"
)

COLUMNS_TWO_FMT = "{0:<20}: {1:<20}"

COLUMNS_DEFAULT = [("all", "*")]
COLUMNS_VM_MIN = [("uuid",), ("name",)]
COLUMNS_VIM_REQUEST = [("vm_uuid", "vm_uuid"), ("vm_name", "vm_name")]
COLUMNS_MOREF = [("moref",), ("name",)]
COLUMNS_FOLDER_MIN = [*COLUMNS_MOREF, ("path",), ("parent.name",)]
COLUMNS_FOLDER = [*COLUMNS_FOLDER_MIN, ("parent.moref",), ("has_children",)]
COLUMNS_NET_MIN = [
    *COLUMNS_MOREF,
    ("description",),
    ("subnet",),
    ("vlan_id",),
    ("vms",),
]
COLUMNS_NET = [
    *COLUMNS_NET_MIN,
    ("ports",),
    ("admin",),
    ("client",),
    ("updated_on",),
]
COLUMNS_PERMISSION = [("principal",), ("group",), ("propagate",)]
COLUMNS_MIN = [("id",), ("created_on",), ("updated_on",)]
COLUMNS_VSS_SERVICE = [("id",), ("label",), ("name",), ("group.name",)]
COLUMNS_IMAGE = [("id",), ("path",), ("name",)]
COLUMNS_OS = [("id",), ("guest_id",), ("full_name",), ("family",)]
COLUMNS_REQUEST = [*COLUMNS_MIN, ("status",)]
COLUMNS_REQUEST_WAIT = [('warnings', 'warnings[*]'), ('errors', 'errors[*]')]
COLUMNS_REQUEST_MAX = [
    ("errors", "message.errors[*]"),
    ("warnings", "message.warnings[*]"),
    ("task_id", "task_id"),
    ("user.username",),
]
COLUMNS_REQUEST_IMAGE_SYNC_MIN = [*COLUMNS_REQUEST, ("type",)]
COLUMNS_REQUEST_IMAGE_SYNC = [
    *COLUMNS_REQUEST,
    ("type",),
    ("deleted",),
    ("added",),
    *COLUMNS_REQUEST_MAX,
]
COLUMNS_REQUEST_SUBMITTED = [
    ("id", "request.id"),
    ("status", "request.status"),
    ("task_id", "request.task_id"),
    ("message",),
]
COLUMNS_REQUEST_MULT_SUBMITTED = [
    ("id", "request.id[*]"),
    ("status", "request.status[*]"),
    ("task_id", "request.task_id[*]"),
    ("message",),
]
COLUMNS_REQUEST_SNAP = [
    ("snapshot.from_date",),
    ("snapshot.to_date",),
    ("snapshot.description",),
    ("snapshot.snap_id",),
    ("extensions",),
    ("action",),
    *COLUMNS_VIM_REQUEST,
    *COLUMNS_REQUEST_MAX,
]
COLUMNS_REQUEST_CHANGE_MIN = [
    *COLUMNS_REQUEST,
    *COLUMNS_VIM_REQUEST,
    ("approval.approved",),
    ("attribute",),
]
COLUMNS_REQUEST_CHANGE = [
    *COLUMNS_REQUEST_CHANGE_MIN,
    ("value", "value[*]"),
    ("scheduled_datetime",),
    *COLUMNS_REQUEST_MAX,
]
COLUMNS_REQUEST_EXPORT_MIN = [
    *COLUMNS_REQUEST,
    *COLUMNS_VIM_REQUEST,
    ("transferred",),
]
COLUMNS_REQUEST_EXPORT = [
    *COLUMNS_REQUEST_EXPORT_MIN,
    ("files", "files[*]"),
    *COLUMNS_REQUEST_MAX,
]
COLUMNS_REQUEST_FOLDER_MIN = [*COLUMNS_REQUEST, ("action",), ("moref",)]
COLUMNS_REQUEST_FOLDER = [*COLUMNS_REQUEST_FOLDER_MIN, *COLUMNS_REQUEST_MAX]
COLUMNS_REQUEST_INVENTORY_MIN = [
    *COLUMNS_REQUEST,
    ("name",),
    ("format",),
    ("transfer",),
]
COLUMNS_REQUEST_INVENTORY = [
    *COLUMNS_REQUEST_INVENTORY_MIN,
    ("properties", "properties.data[*]"),
    ("filters",),
    ("transfer",),
    ("transferred",),
    *COLUMNS_REQUEST_MAX,
]
COLUMNS_REQUEST_NEW_MIN = [
    *COLUMNS_REQUEST,
    *COLUMNS_VIM_REQUEST,
    ("approval.approved",),
    ("built_from",),
]
COLUMNS_REQUEST_NEW = [
    *COLUMNS_REQUEST_NEW_MIN,
    ("domain",),
    ("source_vm",),
    ("source_template",),
    ("source_image",),
    ("folder",),
    ("cpu",),
    ("memory",),
    ("disks", "disks[*]"),
    ("networks", "networks[*]"),
    *COLUMNS_REQUEST_MAX,
]
COLUMNS_TK_MIN = [
    ("id",),
    ("created_on",),
    ("updated_on",),
    ("last_access",),
    ("ip_address",),
    ("valid",),
]
COLUMNS_TK = [*COLUMNS_TK_MIN, ("type",), ("expiration",), ("duration",)]
COLUMNS_MESSAGE_MIN = [*COLUMNS_MIN, ("kind",), ("subject",), ("status",)]
COLUMNS_MESSAGE = [
    *COLUMNS_MIN,
    ("kind",),
    ("status",),
    ("user.username",),
    ("subject",),
    ("text",),
]
COLUMNS_VM_TEMPLATE = [
    *COLUMNS_VM_MIN,
    ("folder.path",),
    ("cpu_count",),
    ("memory_gb",),
]
COLUMNS_VM = [*COLUMNS_VM_TEMPLATE, ("power_state",), ("ip_address",)]
COLUMNS_VM_INFO = [
    ("uuid",),
    ("name", "name.full_name"),
    ("folder.path", "folder.path"),
    ("guest_id", "config.os.guest_id"),
    ("version", "hardware.version"),
    ("overall_status", "state.overall_status"),
    ("power_state", "state.power_state"),
    ("alarms", "state.alarms"),
    ("cpu", "hardware.cpu.cpu_count"),
    ("memory_gb", "hardware.memory.memory_gb"),
    ("provisioned_gb", "storage.provisioned_gb"),
    ("snapshot", "snapshot.exist"),
    ("disks", "hardware.devices.disks[*].unit"),
    ("nics", "hardware.devices.nics[*].unit"),
    ("floppies", "hardware.devices.floppies[*].unit"),
]
COLUMNS_VM_GUEST = [
    ("hostname",),
    ("ip_address", "ip_address[*]"),
    ("full_name", "os.full_name"),
    ("guest_id", "os.guest_id"),
    ("running_status", "tools.running_status"),
]
COLUMNS_VM_GUEST_OS = [
    ("family", "guest_family"),
    ("full_name", "guest_full_name"),
    ("id", "guest_id"),
]
COLUMNS_VM_GUEST_IP = [
    ("ip_address", "ip_address"),
    ("mac_address", "mac_address"),
    ("origin",),
    ("state",),
]
COLUMNS_VM_HAGROUP = [*COLUMNS_VM_MIN, ("VALID", "valid")]
COLUMNS_VM_MEMORY = [
    ("memory_gb",),
    ("hot_add.enabled",),
    ("hot_add.limit_gb",),
    ("ballooned_memory_mb", "quick_stats.ballooned_memory_mb"),
    ("guest_memory_usage_mb", "quick_stats.guest_memory_usage_mb"),
]
COLUMNS_VM_NIC_MIN = [
    ("label",),
    ("mac_address",),
    ("type",),
    ("network.name",),
    ("network.moref",),
    ("connected",),
]
COLUMNS_VM_NIC = [*COLUMNS_VM_NIC_MIN, ("start_connected",)]
COLUMNS_OBJ_PERMISSION = [("principal",), ("group",), ("propagate",)]
COLUMNS_VM_SNAP_MIN = [("id",), ("name",)]
COLUMNS_VM_SNAP = [
    *COLUMNS_VM_SNAP_MIN,
    ("power_state",),
    ("quiesced",),
    ("size_gb",),
    ("description",),
    ("create_time",),
    ("age",),
]
COLUMNS_VM_ADMIN = [("name",), ("email",), ("phone",)]
COLUMNS_VM_ALARM_MIN = [*COLUMNS_MOREF, ("overall_status",), ("date_time",)]
COLUMNS_VM_ALARM = [
    *COLUMNS_VM_ALARM_MIN,
    ("acknowledged",),
    ("acknowledged_by_user",),
    ("acknowledged_date_time",),
]
COLUMNS_VM_BOOT = [
    ("enter_bios_setup",),
    ("boot_retry_delay_ms",),
    ("boot_delay_ms",),
]
COLUMNS_VM_CD_MIN = [("label",), ("backing",), ("connected",)]
COLUMNS_VM_CD = [
    *COLUMNS_VM_CD_MIN,
    ("controller.type",),
    ("controller.virtual_device_node",),
]
COLUMNS_VM_CTRL_MIN = [("label",), ("bus_number",), ("type",)]
COLUMNS_VM_CTRL = [
    *COLUMNS_VM_CTRL_MIN,
    ("controller_key",),
    ("summary",),
    ("shared_bus",),
    ("hot_add_rmove",),
]
COLUMNS_VM_DISK_MIN = [
    ("label",),
    ("unit",),
    ("controller.virtual_device_node",),
]
COLUMNS_VM_DISK = [*COLUMNS_VM_DISK_MIN, ("capacity_gb",), ("shares.level",)]

COLUMNS_VM_DISK_BACKING = [
    ("descriptor_file_name",),
    ("device_name",),
    ("disk_mode",),
    ("file_name",),
    ("lun_uuid",),
    ("thin_provisioned",),
]
COLUMNS_VM_DISK_SCSI = [("bus_number",), ("label",), ("type",)]
COLUMNS_VM_CTRL_DISK = [
    ("controller.virtual_device_node",),
    *COLUMNS_VM_DISK_MIN,
    ("capacity_gb",),
]
COLUMNS_VM_CPU = [
    ("cpu",),
    ("cores_per_socket",),
    ("hot_add.enabled",),
    ("hot_remove.enabled",),
    ("overall_cpu_demand_mhz", "quick_stats.overall_cpu_demand_mhz"),
    ("overall_cpu_usage_mhz", "quick_stats.overall_cpu_usage_mhz"),
]
COLUMNS_VM_EVENT = [("user_name",), ("created_time",), ("message",)]
COLUMNS_VM_STATE = [
    ("create_date", "state.create_date"),
    ("power_state", "state.power_state"),
    ("boot_time", "state.boot_time"),
    ("connection_state", "state.connection_state"),
    ("domain_name", "domain.name"),
    ("domain_moref", "domain.moref"),
]
COLUMNS_VM_TOOLS = [("version",), ("version_status",), ("running_status",)]
COLUMNS_VM_HW = [
    ("value",),
    ("status",),
    ("upgrade_policy", "upgrade_policy.upgrade_policy"),
]
COLUMNS_VM_CONSOLIDATION = [("require_disk_consolidation",)]
COLUMNS_VM_CONTROLLERS = [('scsi.count',)]
COLUMNS_EXTRA_CONFIG = [("key",), ("value",)]
COLUMNS_VSS_OPTIONS = [("options", "[*]")]
COLUMNS_GROUP = [
    ("cn",),
    ("description",),
    ("create_timestamp",),
    ("modify_timestamp",),
    ("unique_member_count",),
    ("unique_member", "unique_member[*].uid"),
]
COLUMNS_GROUPS = [("groups", "groups[*]")]
COLUMNS_ROLE = [
    ("name",),
    ("description",),
    ("entitlements", "entitlements[*]"),
]
COLUMNS_USER_PERSONAL = [
    ("username",),
    ("full_name",),
    ("email",),
    ("phone",),
    ("auth_timestamp",),
    ("pwd_change_time",),
    ("pwd_account_locked_time",),
]
COLUMNS_USER_STATUS = [
    ("created_on",),
    ("updated_on",),
    ("last_access",),
    ("ip_address",),
]
COLUMNS_MESSAGE_DIGEST = [("message",)]
COLUMNS_NOT_REQUEST = [
    ("all",),
    ("none",),
    ("completion",),
    ("error",),
    ("submission",),
]
COLUMNS_WEBDAV = [("files", "[*]")]
COLUMNS_WEBDAV_INFO = [("created",), ("modified",), ("name",), ("size",)]
COLUMNS_SSH_KEY_MIN = [*COLUMNS_MIN, ("type",), ("comment",)]
COLUMNS_SSH_KEY = [*COLUMNS_SSH_KEY_MIN, ("fingerprint",), ("value",)]
COLUMNS_VMRC = [("enabled",), ("options",)]



import logging

import requests

from vss_cli.utils.emoji import EMOJI_UNICODE
from vss_cli.vssconst import HEALTHCHECK_IO

hc_io_badge = f"https://healthchecks.io/badge/{HEALTHCHECK_IO}"

_LOGGING = logging.getLogger(__name__)


def check_status():
    status = 'unknown'
    icon = EMOJI_UNICODE.get(':question_mark:')
    try:
        r = requests.get(hc_io_badge)
        _status = r.json()
        status = _status.get('status')
        if status in ['up']:
            status = 'operational'
            icon = EMOJI_UNICODE.get(':white_heavy_check_mark:')
        elif status in ['down']:
            icon = EMOJI_UNICODE.get(':cross_mark:')
    except Exception as ex:
        _LOGGING.warning(f'healthchecks.io lookup failed: {ex}')
    return {"name": "ITS Private Cloud API", "status": status, "icon": icon}



"""Details for the auto-completion."""
import os
from typing import Callable, Dict, List, Tuple  # NOQA

from requests.exceptions import HTTPError

from vss_cli import const
from vss_cli.config import Configuration
from vss_cli.exceptions import VssError


def _init_ctx(ctx: Configuration) -> None:
    """Initialize ctx."""
    # ctx is incomplete thus need to 'hack' around it
    # see bug https://github.com/pallets/click/issues/942
    ctx.client = Configuration(tk=os.environ.get('VSS_TOKEN'))
    ctx.client.endpoint = os.environ.get('VSS_ENDPOINT', None)
    ctx.client.username = os.environ.get('VSS_USER', None)
    ctx.client.password = os.environ.get('VSS_USER_PASS', None)
    ctx.client.timeout = int(
        os.environ.get('VSS_TIMEOUT', str(const.DEFAULT_TIMEOUT))
    )
    ctx.client.config = os.environ.get('VSS_CONFIG', const.DEFAULT_CONFIG)
    # fallback to load configuration
    ctx.client.load_config()


def _autocomplete(
    f: Callable,
    incomplete: str,
    attrs: List[str],
    f_kwargs: Dict = None,
    sort_index: int = 0,
    complete_index: int = 0,
) -> List[Tuple[str, str]]:
    try:
        response = f(**f_kwargs)
    except (HTTPError, VssError):
        response = []

    completions = []  # type: List[Tuple[str, str]]
    if response:
        for obj in response:
            sub_attrs = []
            # start iterating in second item (index 1)
            for attr in attrs[1:]:
                sub_attrs.append(str(obj.get(attr, 'N/A')))
            # first item (index 0) is always the key
            r = (str(obj[attrs[0]]), ' - '.join(sub_attrs))
            completions.append(r)

        completions.sort(key=lambda x: x[sort_index])

        return [c for c in completions if incomplete in c[complete_index]]
    return completions


def table_formats(
    ctx: Configuration, args: List, incomplete: str
) -> List[Tuple[str, str]]:
    """Table Formats."""
    _init_ctx(ctx)

    completions = [
        ("plain", "Plain tables, no pseudo-graphics to draw lines"),
        ("simple", "Simple table with --- as header/footer (default)"),
        ("github", "Github flavored Markdown table"),
        ("grid", "Formatted as Emacs 'table.el' package"),
        ("fancy_grid", "Draws a fancy grid using box-drawing characters"),
        ("pipe", "PHP Markdown Extra"),
        ("orgtbl", "org-mode table"),
        ("jira", "Atlassian Jira Markup"),
        ("presto", "Formatted as PrestoDB cli"),
        ("psql", "Formatted as Postgres psql cli"),
        ("rst", "reStructuredText"),
        ("mediawiki", "Media Wiki as used in Wikpedia"),
        ("moinmoin", "MoinMain Wiki"),
        ("youtrack", "Youtrack format"),
        ("html", "HTML Markup"),
        ("latex", "LaTeX markup, replacing special characters"),
        ("latex_raw", "LaTeX markup, no replacing of special characters"),
        (
            "latex_booktabs",
            "LaTex markup using spacing and style from `booktabs",
        ),
        ("textile", "Textile"),
        ("tsv", "Tab Separated Values"),
        ("csv", "Comma Separated Values"),
    ]

    completions.sort()

    return [c for c in completions if incomplete in c[0]]


def vm_templates(
    ctx: Configuration, args: List, incomplete: str
) -> List[Tuple[str, str]]:
    _init_ctx(ctx)
    return _autocomplete(
        ctx.client.get_templates,
        incomplete,
        ['uuid', 'name'],
        complete_index=1,
        sort_index=1,
        f_kwargs={"short": 1, "show_all": True, "per_page": 2000},
    )


def virtual_machines(
    ctx: Configuration, args: List, incomplete: str
) -> List[Tuple[str, str]]:
    _init_ctx(ctx)
    return _autocomplete(
        ctx.client.get_vms,
        incomplete,
        ['uuid', 'name'],
        complete_index=1,
        sort_index=1,
        f_kwargs={"show_all": True, "short": 1, "per_page": 2000},
    )


def vm_controller_scsi_types(
    ctx: Configuration, args: List, incomplete: str
) -> List[Tuple[str, str]]:
    _init_ctx(ctx)
    return _autocomplete(
        ctx.client.get_supported_scsi_controllers,
        incomplete,
        ['type', 'description'],
        f_kwargs={'only_type': False},
    )


def vm_disk_backing_modes(
    ctx: Configuration, args: List, incomplete: str
) -> List[Tuple[str, str]]:
    _init_ctx(ctx)
    return _autocomplete(
        ctx.client.get_supported_disk_backing_modes,
        incomplete,
        ['type', 'description'],
        f_kwargs={'only_type': False},
    )


def domains(
    ctx: Configuration, args: List, incomplete: str
) -> List[Tuple[str, str]]:
    _init_ctx(ctx)
    return _autocomplete(
        ctx.client.get_domains,
        incomplete,
        ['moref', 'name'],
        complete_index=1,
        sort_index=1,
        f_kwargs={"short": 1, "show_all": True, "per_page": 2000},
    )


def folders(
    ctx: Configuration, args: List, incomplete: str
) -> List[Tuple[str, str]]:
    _init_ctx(ctx)
    return _autocomplete(
        ctx.client.get_folders,
        incomplete,
        ['moref', 'path'],
        sort_index=1,
        complete_index=1,
        f_kwargs={
            "show_all": True,
            "sort": "path,desc",
            "per_page": 500,
            "short": 1,
        },
    )


def networks(
    ctx: Configuration, args: List, incomplete: str
) -> List[Tuple[str, str]]:
    _init_ctx(ctx)
    return _autocomplete(
        ctx.client.get_networks,
        incomplete,
        ['moref', 'label'],
        sort_index=1,
        complete_index=1,
        f_kwargs={"show_all": True, "sort": "label,desc", "per_page": 500},
    )


def operating_systems(
    ctx: Configuration, args: List, incomplete: str
) -> List[Tuple[str, str]]:
    _init_ctx(ctx)
    return _autocomplete(
        ctx.client.get_os,
        incomplete,
        ['guest_id', 'full_name'],
        f_kwargs={"show_all": True, "sort": "guest_id,desc", "per_page": 500},
    )


def vss_services(
    ctx: Configuration, args: List, incomplete: str
) -> List[Tuple[str, str]]:
    _init_ctx(ctx)
    return _autocomplete(
        ctx.client.get_vss_services,
        incomplete,
        ['id', 'label'],
        sort_index=1,
        complete_index=1,
        f_kwargs={"show_all": True, "sort": "label,desc", "per_page": 500},
    )


def isos(
    ctx: Configuration, args: List, incomplete: str
) -> List[Tuple[str, str]]:
    _init_ctx(ctx)
    return _autocomplete(
        ctx.client.get_isos,
        incomplete,
        attrs=['id', 'path'],
        sort_index=1,
        complete_index=1,
        f_kwargs={"show_all": True, "sort": "path,desc", "per_page": 500},
    )


def vm_images(
    ctx: Configuration, args: List, incomplete: str
) -> List[Tuple[str, str]]:
    _init_ctx(ctx)
    return _autocomplete(
        ctx.client.get_images,
        incomplete,
        attrs=['id', 'path'],
        sort_index=1,
        complete_index=1,
        f_kwargs={"show_all": True, "sort": "path,desc", "per_page": 500},
    )


def floppies(
    ctx: Configuration, args: List, incomplete: str
) -> List[Tuple[str, str]]:
    _init_ctx(ctx)
    return _autocomplete(
        ctx.client.get_floppies,
        incomplete,
        attrs=['id', 'path'],
        sort_index=1,
        complete_index=1,
        f_kwargs={"show_all": True, "sort": "path,desc", "per_page": 500},
    )


def inventory_properties(
    ctx: Configuration, args: List, incomplete: str
) -> List[Tuple[str, str]]:
    _init_ctx(ctx)
    try:
        response = ctx.client.request('/inventory/options')
        if response:
            response = response.get('data')
    except (HTTPError, VssError):
        response = []

    completions = []  # type: List[Tuple[str, str]]
    if response:
        for obj in response:
            completions.append((obj['key'], obj['value']))

        completions.sort()

        return [c for c in completions if incomplete in c[0]]
    return completions


def inventory_requests(
    ctx: Configuration, args: List, incomplete: str
) -> List[Tuple[str, str]]:
    _init_ctx(ctx)
    return _autocomplete(
        ctx.client.get_inventory_requests,
        incomplete,
        attrs=['id', 'created_on', 'name'],
        f_kwargs={"sort": "created_on,desc", "per_page": 500},
    )


def change_requests(
    ctx: Configuration, args: List, incomplete: str
) -> List[Tuple[str, str]]:
    _init_ctx(ctx)
    return _autocomplete(
        ctx.client.get_change_requests,
        incomplete,
        attrs=['id', 'vm_uuid', 'vm_name', 'attribute'],
        f_kwargs={"sort": "created_on,desc", "per_page": 500},
    )


def export_requests(
    ctx: Configuration, args: List, incomplete: str
) -> List[Tuple[str, str]]:
    _init_ctx(ctx)
    return _autocomplete(
        ctx.client.get_export_requests,
        incomplete,
        attrs=['id', 'vm_uuid', 'vm_name'],
        f_kwargs={"sort": "created_on,desc", "per_page": 500},
    )


def folder_requests(
    ctx: Configuration, args: List, incomplete: str
) -> List[Tuple[str, str]]:
    _init_ctx(ctx)
    return _autocomplete(
        ctx.client.get_folder_requests,
        incomplete,
        attrs=['id', 'moref', 'action'],
        f_kwargs={"sort": "created_on,desc", "per_page": 500},
    )


def image_sync_requests(
    ctx: Configuration, args: List, incomplete: str
) -> List[Tuple[str, str]]:
    _init_ctx(ctx)
    return _autocomplete(
        ctx.client.get_folder_requests,
        incomplete,
        attrs=['id', 'type'],
        f_kwargs={"sort": "created_on,desc", "per_page": 500},
    )


def snapshot_requests(
    ctx: Configuration, args: List, incomplete: str
) -> List[Tuple[str, str]]:
    _init_ctx(ctx)
    return _autocomplete(
        ctx.client.get_snapshot_requests,
        incomplete,
        attrs=['id', 'vm_uuid', 'vm_name'],
        f_kwargs={"sort": "created_on,desc", "per_page": 500},
    )


def account_messages(
    ctx: Configuration, args: List, incomplete: str
) -> List[Tuple[str, str]]:
    _init_ctx(ctx)
    return _autocomplete(
        ctx.client.get_user_messages,
        incomplete,
        attrs=['id', 'kind', 'subject'],
        f_kwargs={"sort": "created_on,desc", "per_page": 500},
    )


def virtual_nic_types(
    ctx: Configuration, args: List, incomplete: str
) -> List[Tuple[str, str]]:
    _init_ctx(ctx)
    return _autocomplete(
        ctx.client.get_supported_nic_types,
        incomplete,
        attrs=['type', 'description'],
        f_kwargs={"only_type": False},
    )


def virtual_hw_types(
    ctx: Configuration, args: List, incomplete: str
) -> List[Tuple[str, str]]:
    _init_ctx(ctx)
    return _autocomplete(
        ctx.client.get_supported_vmx_types,
        incomplete,
        attrs=['type', 'description'],
        f_kwargs={"only_type": False},
    )


def vss_options(
    ctx: Configuration, args: List, incomplete: str
) -> List[Tuple[str, str]]:
    _init_ctx(ctx)
    return _autocomplete(
        ctx.client.get_supported_vss_options,
        incomplete,
        attrs=['option', 'description'],
        f_kwargs={"only_option": False},
    )



"""Token Management plugin for VSS CLI (vss-cli)."""
import click

from vss_cli import const, rel_opts as so
from vss_cli.cli import pass_context
from vss_cli.config import Configuration
from vss_cli.helper import format_output


@click.group('token', short_help='Manage access tokens')
@pass_context
def cli(ctx: Configuration):
    """Manage access tokens."""
    with ctx.spinner(disable=ctx.debug):
        ctx.load_config()


@cli.command('ls', short_help='list user tokens')
@so.filter_opt
@so.sort_opt
@so.all_opt
@so.count_opt
@so.page_opt
@pass_context
def token_ls(ctx: Configuration, filter_by, page, sort, show_all, count):
    """List tokens based on:

        Filter list in the following format <field_name>=<operator>,<value>
        where operator is eq, ne, lt, le, gt, ge, like, in.
        For example: valid=eq,false

            vss-cli token ls -f valid=eq,false

        Sort list in the following format <field_name>=<asc|desc>. For example:

            vss-cli token ls -s created_on=desc

    """
    columns = ctx.columns or const.COLUMNS_TK_MIN
    params = dict()
    if all(filter_by):
        params['filter'] = ';'.join(filter_by)
    if all(sort):
        params['sort'] = ';'.join(sort)
    # make request
    with ctx.spinner(disable=ctx.debug):
        _requests = ctx.get_user_tokens(
            show_all=show_all, per_page=count, **params
        )
    # format output
    output = format_output(ctx, _requests, columns=columns)
    # page results
    if page:
        click.echo_via_pager(output)
    else:
        click.echo(output)


@cli.command('get', help='Token')
@click.argument('tid', type=click.INT, required=True)
@pass_context
def token_get(ctx: Configuration, tid):
    with ctx.spinner(disable=ctx.debug):
        obj = ctx.get_user_token(tid)
    columns = ctx.columns or const.COLUMNS_TK
    click.echo(format_output(ctx, [obj], columns=columns, single=True))


@cli.command('rm', help='Delete user token.')
@click.argument('tid', type=int, required=True, nargs=-1)
@click.option('-s', '--summary', is_flag=True, help='Print request summary')
@pass_context
def token_rm(ctx: Configuration, tid, summary):
    result = []
    with click.progressbar(tid) as ids:
        for i in ids:
            result.append(ctx.delete_user_token(i))
    if summary:
        for res in result:
            click.echo(
                format_output(
                    ctx, [res], columns=[('STATUS', 'status')], single=True
                )
            )



"""Inventory Request Management plugin for VSS CLI (vss-cli)."""
import logging

import click

from vss_cli import const, rel_opts as so
import vss_cli.autocompletion as autocompletion
from vss_cli.cli import pass_context
from vss_cli.config import Configuration
from vss_cli.helper import format_output
from vss_cli.plugins.request import cli

_LOGGING = logging.getLogger(__name__)


@cli.group('inventory', short_help='Manage virtual machine inventory requests')
@pass_context
def request_mgmt_inventory(ctx: Configuration):
    """Manage virtual machine inventory requests"""


@request_mgmt_inventory.command('ls', short_help='list inventory requests')
@so.filter_opt
@so.sort_opt
@so.all_opt
@so.count_opt
@so.page_opt
@pass_context
def request_mgmt_inventory_ls(
    ctx: Configuration, filter_by, page, sort, show_all, count
):
    """List requests based on:

        Filter list in the following format <field_name> <operator>,<value>
        where operator is eq, ne, lt, le, gt, ge, like, in.
        For example: status=eq,PROCESSED

            vss-cli request inventory ls -f status=eq,PROCESSED

        Sort list in the following format <field_name>=<asc|desc>. For example:

            vss-cli request inventory ls -s created_on=desc

    """
    columns = ctx.columns or const.COLUMNS_REQUEST_INVENTORY_MIN
    params = dict(expand=1, sort='created_on,desc')
    if all(filter_by):
        params['filter'] = ';'.join(filter_by)
    if all(sort):
        params['sort'] = ';'.join(sort)
    # make request
    with ctx.spinner(disable=ctx.debug):
        _requests = ctx.get_inventory_requests(
            show_all=show_all, per_page=count, **params
        )

    output = format_output(ctx, _requests, columns=columns)
    # page results
    if page:
        click.echo_via_pager(output)
    else:
        click.echo(output)


@request_mgmt_inventory.command('get', short_help='Inventory request')
@click.argument(
    'rid',
    type=click.INT,
    required=True,
    autocompletion=autocompletion.inventory_requests,
)
@pass_context
def request_mgmt_inventory_get(ctx: Configuration, rid):
    # make request
    with ctx.spinner(disable=ctx.debug):
        obj = ctx.get_inventory_request(rid)
    columns = ctx.columns or const.COLUMNS_REQUEST_INVENTORY
    click.echo(format_output(ctx, [obj], columns=columns, single=True))



"""Raw plugin for VSS CLI (vss-cli)."""
import json as json_
import logging

import click
from vss_cli.cli import pass_context
from vss_cli.config import Configuration
from vss_cli.helper import format_output

_LOGGING = logging.getLogger(__name__)


@click.command(
    'raw',
    short_help='Make a raw call to the API'
)
@click.argument(
    'method',
    default='GET',
    type=click.Choice(
        ['GET', 'POST', 'PUT',
         'DELETE', 'PATCH']
    )
)
@click.argument(
    'uri',
)
@click.option(
    '--json', '-j',
    help='JSON payload',
    default=None
)
@pass_context
def cli(ctx: Configuration, method, uri, json):
    """Call the raw API (advanced)."""
    ctx.load_config()
    ctx.auto_output("data")
    req = dict(url=uri, method=method.upper())
    if json:
        data = json_.loads(json)
        req['payload'] = data
    # make request
    with ctx.spinner(disable=ctx.debug):
        obj = ctx.request(**req)
    ctx.echo(
        format_output(
            ctx,
            obj
        )
    )



from dataclasses import dataclass, field
import re
from typing import List, Optional, Union

from dataclasses_json import dataclass_json
from validators import url as is_url

import vss_cli.const as const
from vss_cli.exceptions import ValidationError


class Url(str):
    def __new__(cls, val):
        if is_url(val):
            return str.__new__(cls, val)
        else:
            raise ValidationError(f'{val} is not a valid URL')


@dataclass_json
@dataclass
class ConfigFileGeneral:
    check_for_updates: bool = const.DEFAULT_CHECK_UPDATES
    check_for_messages: bool = const.DEFAULT_CHECK_MESSAGES
    default_endpoint_name: str = const.DEFAULT_ENDPOINT_NAME
    debug: bool = False
    output: str = const.DEFAULT_RAW_OUTPUT
    table_format: str = 'simple'
    timeout: int = const.DEFAULT_TIMEOUT
    verbose: bool = False
    columns_width: int = -1


@dataclass_json
@dataclass
class ConfigEndpoint:
    url: Url
    name: Optional[str] = None
    auth: Optional[str] = None
    token: Optional[str] = None

    def __post_init__(self):
        def get_hostname_from_url(
            url: str, hostname_regex: str = const.DEFAULT_HOST_REGEX
        ) -> str:
            """Parse hostname from URL"""
            re_search = re.search(hostname_regex, url)
            _, _hostname = re_search.groups() if re_search else ('', '')
            _host = _hostname.split('.')[0] if _hostname.split('.') else ''
            return _host

        if not self.name:
            self.name = get_hostname_from_url(self.url)


@dataclass_json
@dataclass
class ConfigFile:
    general: ConfigFileGeneral
    endpoints: Optional[Union[List[ConfigEndpoint]]] = field(
        default_factory=lambda: []
    )

    def get_endpoint(self, ep_name_or_url: str) -> List[ConfigEndpoint]:
        if self.endpoints:
            ep = list(
                filter(lambda i: ep_name_or_url == i.name, self.endpoints)
            ) or list(
                filter(lambda i: ep_name_or_url == i.url, self.endpoints)
            )
            return ep
        else:
            return []

    def update_endpoint(
        self, endpoint: ConfigEndpoint
    ) -> List[ConfigEndpoint]:
        if self.endpoints:
            for idx, val in enumerate(self.endpoints):
                if val.name == endpoint.name:
                    self.endpoints[idx] = endpoint
                    return self.endpoints
        else:
            self.endpoints = []
        # adding
        self.endpoints.append(endpoint)
        return self.endpoints

    def update_endpoints(
        self, *endpoints: List[ConfigEndpoint]
    ) -> List[ConfigEndpoint]:
        for endpoint in endpoints:
            self.update_endpoint(endpoint)
        return self.endpoints



"""VSS Storage Management plugin for VSS CLI (vss-cli)."""
import logging
import os

import click
from vss_cli import const
from vss_cli.cli import pass_context
from vss_cli.config import Configuration
from vss_cli.helper import format_output
from vss_cli.utils.emoji import EMOJI_UNICODE

_LOGGING = logging.getLogger(__name__)


@click.group(
    'stor',
    short_help='Manage your VSS storage account.'
)
@pass_context
def cli(ctx: Configuration):
    """Manage your VSS storage account."""
    with ctx.spinner(disable=ctx.debug):
        ctx.load_config()


@cli.command(
    'ls',
    short_help='list remote dir contents'
)
@click.argument(
    'remote_path',
    type=click.STRING,
    default="/"
)
@pass_context
def stor_ls(ctx: Configuration, remote_path):
    columns = ctx.columns or const.COLUMNS_WEBDAV
    ctx.get_vskey_stor()
    with ctx.spinner(disable=ctx.debug):
        obj = ctx.vskey_stor.list(remote_path)
    click.echo(
        format_output(
            ctx,
            obj,
            columns=columns
        )
    )


@cli.command(
    'get',
    short_help='get remote info'
)
@click.argument(
    'remote_path',
    type=click.STRING,
    required=True
)
@pass_context
def stor_get(ctx, remote_path):
    ctx.get_vskey_stor()
    columns = ctx.columns or const.COLUMNS_WEBDAV_INFO
    obj = ctx.vskey_stor.info(remote_path)
    click.echo(
        format_output(
            ctx,
            [obj],
            columns=columns,
            single=True
        )
    )


@cli.command(
    'dl',
    short_help='download file'
)
@click.argument(
    'remote_path',
    type=click.STRING,
    required=True
)
@click.option(
    '-d', '--dir',
    type=click.STRING,
    help='Local target directory',
    default=os.getcwd()
)
@click.option(
    '-n', '--name',
    type=click.STRING,
    help='Local target name'
)
@pass_context
def stor_dl(
        ctx: Configuration,
        remote_path, dir, name
):
    """Download remote file."""
    ctx.get_vskey_stor()
    local_dir = os.path.expanduser(dir) or os.getcwd()
    local_name = name or os.path.basename(remote_path)
    local_path = os.path.join(local_dir, local_name)
    # check if remote path exists
    if not ctx.vskey_stor.check(remote_path):
        raise _LOGGING.warning(f'Remote path not found {remote_path}')
    ctx.log(
        f'Download {remote_path} to {local_path} in '
        f'progress {EMOJI_UNICODE.get(":fast_down_button:")} '
    )
    with ctx.spinner(disable=ctx.debug):
        ctx.vskey_stor.download_sync(
            remote_path=remote_path,
            local_path=local_path
        )
    ctx.log(
        f'Download complete to {local_path} '
        f'{EMOJI_UNICODE.get(":white_heavy_check_mark:")}')


@cli.command(
    'ul',
    short_help='upload file'
)
@click.argument(
    'file_path',
    type=click.Path(exists=True),
    required=True
)
@click.option(
    '-d', '--dir',
    type=click.STRING,
    help='Remote target directory',
    default='/'
)
@click.option(
    '-n', '--name',
    type=click.STRING,
    help='Remote target name'
)
@pass_context
def stor_ul(ctx: Configuration, file_path, name, dir):
    """Upload given file to your VSKEY-STOR space.
    This command is useful when, for instance, a required ISO is
    not available in the VSS central repository and needs to be
    mounted to a virtual machine.
    """
    ctx.get_vskey_stor()
    file_name = name or os.path.basename(file_path)
    remote_base = dir
    # check if remote path exists
    if not ctx.vskey_stor.check(remote_base):
        _LOGGING.debug(f'Creating {remote_base}.')
        ctx.vskey_stor.mkdir(remote_base)
        _LOGGING.debug(f'Created {remote_base}.')
    # upload
    remote_path = os.path.join(remote_base, file_name)
    ctx.log(f'Upload {file_path} to {remote_path} '
            f'in progress {EMOJI_UNICODE.get(":fast_up_button:")}')
    with ctx.spinner(disable=ctx.debug):
        ctx.vskey_stor.upload_sync(
            remote_path=remote_path,
            local_path=file_path)
    ctx.echo(
        f'Upload complete to {remote_path} '
        f'{EMOJI_UNICODE.get(":white_heavy_check_mark:")}'
    )



import logging

import click

from vss_cli import const, rel_opts as so
from vss_cli.cli import pass_context
from vss_cli.config import Configuration
from vss_cli.helper import format_output
from vss_cli.plugins.compute import cli

_LOGGING = logging.getLogger(__name__)


@cli.group('os', short_help='Supported OS.')
@pass_context
def compute_os(ctx):
    """Supported operating systems by our infrastructure.
    This resource is useful when deploying a new or
    reconfiguring an existing virtual machine."""


@compute_os.command('ls', short_help='list operating systems')
@so.filter_opt
@so.sort_opt
@so.all_opt
@so.count_opt
@so.page_opt
@pass_context
def compute_os_ls(ctx: Configuration, filter_by, page, sort, show_all, count):
    """List requests based on:

        Filter list in the following format <field_name>=<operator>,<value>
        where operator is eq, ne, lt, le, gt, ge, like, in.
        For example: status=eq,PROCESSED

            vss-cli compute os ls -f full_name=like,CentOS%

        Sort list in the following format <field_name>=<asc|desc>. For example:

            vss-cli compute os ls -s guest_id=asc

    """
    params = dict(expand=1, sort='guest_id,asc')
    if all(filter_by):
        params['filter'] = ';'.join(filter_by)
    if all(sort):
        params['sort'] = ';'.join(sort)
    with ctx.spinner(disable=ctx.debug):
        obj = ctx.get_os(show_all=show_all, per_page=count, **params)
    # format
    columns = ctx.columns or const.COLUMNS_OS
    output = format_output(ctx, obj, columns=columns)
    # page results
    if page:
        click.echo_via_pager(output)
    else:
        click.echo(output)



"""Auto-completion for VSS CLI (vss-cli)."""
import click
from click._bashcomplete import get_completion_script

from vss_cli.cli import pass_context


@click.group(
    'completion',
    short_help='Output shell completion code for '
    'the specified shell (bash or zsh or fish).',
)
@pass_context
def cli(ctx):
    """Output shell completion code for the specified
    shell (bash or zsh)."""


def dump_script(shell: str) -> None:
    """Dump the script content."""
    prog_name = "vss-cli"
    cvar = '_%s_COMPLETE' % (prog_name.replace('-', '_')).upper()

    click.echo(get_completion_script(prog_name, cvar, shell))


@cli.command()
@pass_context
def bash(ctx):
    """Output shell completion code for bash."""
    dump_script("bash")


@cli.command()
@pass_context
def zsh(ctx):
    """Output shell completion code for zsh."""
    dump_script("zsh")


@cli.command()
@pass_context
def fish(ctx):
    """Output shell completion code for fish."""
    dump_script("fish")



import logging
from typing import List

import click

from vss_cli import const, rel_opts as so
import vss_cli.autocompletion as autocompletion
from vss_cli.cli import pass_context
from vss_cli.config import Configuration
from vss_cli.helper import format_output
from vss_cli.plugins.compute import cli

_LOGGING = logging.getLogger(__name__)


@cli.group('inventory', short_help='Manage inventory reports')
@pass_context
def cli(ctx: Configuration):
    """Create or download an inventory file of your virtual machines
    hosted. Inventory files are created and transferred to your VSKEY-STOR
    space and are also available through the API."""


@cli.command('dl', short_help='download inventory report')
@click.argument(
    'request_id',
    type=click.INT,
    required=True,
    autocompletion=autocompletion.inventory_requests,
)
@click.option(
    '-d',
    '--directory',
    type=click.STRING,
    help='report destination',
    required=False,
    default=None,
)
@click.option(
    '-l', '--launch', is_flag=True, help='Launch link in default application'
)
@pass_context
def compute_inventory_dl(ctx: Configuration, request_id, directory, launch):
    """Downloads given inventory request to current directory or
    provided path. Also, it's possible to open downloaded file in
    default editor."""
    file_path = ctx.download_inventory_file(request_id, directory)
    # to launch or not
    if launch and file_path:
        click.launch(file_path.get('file'))


@cli.command('mk', short_help='create inventory report')
@click.argument(
    'attribute',
    nargs=-1,
    default=None,
    autocompletion=autocompletion.inventory_properties,
)
@click.option(
    '-f',
    '--fmt',
    type=click.Choice(['json', 'csv']),
    default='csv',
    help='report format',
    show_default=True,
)
@click.option(
    '--transfer/--no-transfer',
    default=False,
    help='Transfer report to personal store',
    show_default=True,
)
@click.option('-a', '--all', is_flag=True, help='include all attributes')
@so.wait_opt
@pass_context
def compute_inventory_mk(
    ctx: Configuration,
    fmt: str,
    all: bool,
    attribute: List[str],
    wait: bool,
    transfer: bool,
):
    """Submits an inventory report request to generate file in JSON or CSV
    of your virtual machines.

    """
    ctx.wait = wait
    attributes = ctx.get_inventory_properties() if all else list(attribute)
    obj = ctx.create_inventory_file(
        fmt=fmt, props=attributes, transfer=transfer
    )
    # format output
    click.echo(
        format_output(
            ctx, [obj], columns=const.COLUMNS_REQUEST_SUBMITTED, single=True
        )
    )
    # wait for request
    if ctx.wait:
        ctx.wait_for_request_to(obj)
        if click.confirm('Would you like to download?'):
            ctx.download_inventory_file(obj['request']['id'], '')



"""Plugins plugin for VSS CLI (vss-cli)."""
import logging

import click
from vss_cli.cli import pass_context
from vss_cli.config import Configuration

from pkg_resources import iter_entry_points
from click_plugins import with_plugins

_LOGGING = logging.getLogger(__name__)


@with_plugins(
    iter_entry_points('vss_cli.contrib.plugins')
)
@click.group(
    'plugins',
    short_help='External plugins.'
)
@pass_context
def cli(ctx: Configuration):
    """External CLI plugins."""
    with ctx.spinner(disable=ctx.debug):
        ctx.load_config()



"""Request Management plugin for VSS CLI (vss-cli)."""
import click
from vss_cli.cli import pass_context
from vss_cli.config import Configuration


@click.group(
    'request',
    short_help='Manage various requests'
)
@pass_context
def cli(ctx: Configuration):
    """Useful to track request status and details."""
    with ctx.spinner(disable=ctx.debug):
        ctx.load_config()


from vss_cli.plugins.request_plugins import (
    snapshot, image, new, change,
    export, folder, inventory
)  # pylint: disable=unused-import



import click

import vss_cli.autocompletion as autocompletion
from vss_cli.plugins.compute_plugins import callbacks
from vss_cli.validators import (
    validate_admin, validate_inform, validate_json_type)

source_opt = click.option(
    '--source',
    '-s',
    help='Source virtual machine or template UUID.',
    type=click.STRING,
    required=True,
    autocompletion=autocompletion.virtual_machines,
)
source_template_opt = click.option(
    '--source',
    '-s',
    help='Source virtual machine or template UUID.',
    type=click.STRING,
    required=True,
    autocompletion=autocompletion.vm_templates,
)
source_image_opt = click.option(
    '--source',
    '-s',
    help='Source Virtual Machine OVA/OVF id, name or path.',
    type=click.STRING,
    required=True,
    autocompletion=autocompletion.vm_images,
)
description_opt = click.option(
    '--description',
    '-d',
    help='A brief description.',
    type=click.STRING,
    required=True,
)
client_nr_opt = click.option(
    '--client',
    '-b',
    help='Client department.',
    type=click.STRING,
    required=False,
)
client_opt = click.option(
    '--client',
    '-b',
    help='Client department.',
    type=click.STRING,
    required=True,
)
admin_opt = click.option(
    '--admin',
    '-a',
    help='Admin name, phone number and email separated by '
    '`:` i.e. "John Doe:416-123-1234:john.doe@utoronto.ca"',
    type=click.STRING,
    callback=validate_admin,
    required=False,
)
inform_opt = click.option(
    '--inform',
    '-r',
    help='Informational contact emails in comma separated',
    type=click.STRING,
    callback=validate_inform,
    required=False,
)
usage_opt = click.option(
    '--usage',
    '-u',
    help='Vm usage.',
    type=click.Choice(['Test', 'Prod', 'Dev', 'QA']),
    required=False,
    default='Test',
)
os_nr_opt = click.option(
    '--os',
    '-o',
    help='Guest operating system id.',
    type=click.STRING,
    required=False,
    autocompletion=autocompletion.operating_systems,
)
os_opt = click.option(
    '--os',
    '-o',
    help='Guest operating system id.',
    type=click.STRING,
    required=True,
    autocompletion=autocompletion.operating_systems,
)
memory_opt = click.option(
    '--memory', '-m', help='Memory in GB.', type=click.INT, required=False
)
cpu_opt = click.option(
    '--cpu', '-c', help='Cpu count.', type=click.INT, required=False
)
folder_nr_opt = click.option(
    '--folder',
    '-f',
    help='Logical folder moref name or path.',
    type=click.STRING,
    required=False,
    autocompletion=autocompletion.folders,
)
folder_opt = click.option(
    '--folder',
    '-f',
    help='Logical folder moref name or path.',
    type=click.STRING,
    required=True,
    autocompletion=autocompletion.folders,
)
disks_nr_opt = click.option(
    '--disk',
    '-i',
    help='Virtual disks in GB.',
    type=click.INT,
    multiple=True,
    required=False,
)
disks_opt = click.option(
    '--disk',
    '-i',
    help='Disks in GB.',
    type=click.INT,
    multiple=True,
    required=True,
)
networks_nr_opt = click.option(
    '--net',
    '-n',
    help='Network adapter <moref-or-name>=<nic-type>.',
    type=click.STRING,
    multiple=True,
    required=False,
    callback=callbacks.process_networks_opt,
    autocompletion=autocompletion.networks,
)
networks_opt = click.option(
    '--net',
    '-n',
    help='Network adapter <moref-or-name>=<nic-type>.',
    type=click.STRING,
    multiple=True,
    required=True,
    callback=callbacks.process_networks_opt,
    autocompletion=autocompletion.networks,
)
domain_opt = click.option(
    '--domain',
    '-t',
    help='Target fault domain name or moref.',
    type=click.STRING,
    required=False,
    autocompletion=autocompletion.domains,
)
notes_opt = click.option(
    '--notes', help='Custom notes.', type=click.STRING, required=False
)
custom_spec_opt = click.option(
    '--custom-spec',
    '-p',
    help='Guest OS custom specification in JSON format.',
    type=click.STRING,
    required=False,
    callback=validate_json_type,
)
iso_opt = click.option(
    '--iso',
    '-s',
    help='ISO image to be mounted after creation',
    type=click.STRING,
    required=False,
    autocompletion=autocompletion.isos,
)
high_io_opt = click.option(
    '--high-io',
    '-h',
    help='Use VMware Paravirtual SCSIController.',
    is_flag=True,
    required=False,
)
power_on_opt = click.option(
    '--power-on',
    help='Power on after successful deployment.',
    is_flag=True,
    required=False,
)
extra_config_opt = click.option(
    '--extra-config',
    '-e',
    help='VMWare Guest Info Interface in JSON format.',
    type=click.STRING,
    required=False,
    callback=validate_json_type,
)
user_data_opt = click.option(
    '--user-data',
    help='Cloud-init user_data YML file path to '
    'pre-configure guest os upon first boot.',
    type=click.File('r'),
    required=False,
)
vss_service_opt = click.option(
    '--vss-service',
    help='VSS Service related to VM',
    autocompletion=autocompletion.vss_services,
    required=False,
)
instances = click.option(
    '--instances',
    help='Number of instances to deploy',
    type=click.INT,
    default=1,
    show_default=True,
)
vss_options_opt = click.option(
    'vss-option',
    help='VSS Option to enable',
    autocompletion=autocompletion.vss_options,
    required=False,
)



"""Misc plugin for VSS CLI (vss-cli)."""
import logging

import click
from vss_cli.cli import pass_context
from vss_cli.config import Configuration


_LOGGING = logging.getLogger(__name__)


@click.group(
    'misc',
    short_help='Miscellaneous utilities.'
)
@pass_context
def cli(ctx):
    """Miscellaneous utilities."""


@cli.command(
    'hash-string',
    short_help='Hashes string using SHA-512'
)
@click.argument(
    'string_to_hash',
    type=click.STRING,
    required=False
)
@pass_context
def misc_hash_string(
        ctx: Configuration, string_to_hash
):
    """Hashes string using SHA-512. If string_to_hash argument
    not provided, input will be prompted. Useful to create hashed
    passwd entries."""
    from pyvss.helper import hash_string
    string_to_hash = string_to_hash or click.prompt(
        'String', show_default=False, hide_input=True,
        confirmation_prompt=True)
    click.echo(
        hash_string(string_to_hash)
    )


@cli.command(
    'gz-b64e',
    short_help='Compresses and encodes a given string'
)
@click.argument(
    'string_gz_encode',
    type=click.STRING,
    required=True
)
@pass_context
def misc_encode_gz(
        ctx: Configuration, string_gz_encode
):
    """Compresses (gz) and encodes in base64 a given string."""
    from pyvss.helper import compress_encode_string
    click.echo(
        compress_encode_string(string_gz_encode)
    )


@cli.command(
    'b64d-gz',
    short_help='Decompress and decodes a given string'
)
@click.argument(
    'string_gz_encoded',
    type=click.STRING, required=True
)
@pass_context
def misc_decodes_ugz(
        ctx: Configuration, string_gz_encoded
):
    """Compresses (gz) and encodes in base64 a given string."""
    from pyvss.helper import decode_uncompress_string
    click.echo(
        decode_uncompress_string(string_gz_encoded)
    )



"""Configuration plugin for VSS CLI (vss-cli)."""
import json
import logging
import os
from typing import Any

import click
from ruamel.yaml.parser import ParserError

from vss_cli import const
from vss_cli.cli import pass_context
from vss_cli.config import Configuration
from vss_cli.data_types import ConfigEndpoint
from vss_cli.helper import format_output, str2bool
from vss_cli.utils.emoji import EMOJI_UNICODE

_LOGGING = logging.getLogger(__name__)


ej_warn = EMOJI_UNICODE.get(':alien:')
ej_rkt = EMOJI_UNICODE.get(':rocket:')
ej_tada = EMOJI_UNICODE.get(':party_popper:')
ej_save = EMOJI_UNICODE.get(':floppy_disk:')
ej_check = EMOJI_UNICODE.get(':white_heavy_check_mark:')


@click.group('configure')
@pass_context
def cli(ctx: Configuration):
    """Configure VSS-CLI options."""
    ctx.auto_output('table')


@cli.command('upgrade', short_help='Upgrade legacy configuration.')
@click.argument(
    'legacy_config',
    envvar='VSS_CONFIG',
    type=click.Path(exists=True),
    default=const.LEGACY_CONFIG,
)
@click.option(
    '-c',
    '--confirm',
    is_flag=True,
    default=False,
    help='Proceed with migration without prompting confirmation.',
)
@click.option(
    '-o',
    '--overwrite',
    is_flag=True,
    default=False,
    help='Overwrite if target file exists.',
)
@pass_context
def upgrade(ctx: Configuration, legacy_config, confirm, overwrite):
    """Upgrade legacy configuration (config.json) to current (config.yaml)."""
    cfg_exc = False
    try:
        with open(legacy_config, 'r') as f:
            legacy_endpoints = ctx.yaml_load(f)
        endpoints = []
        if legacy_endpoints:
            n_ep = len(legacy_endpoints)
            click.echo(
                f'Found {n_ep} endpoints. Migrating to new configuration file.'
            )
            for ep_k, ep_v in legacy_endpoints.items():
                t_ep = {
                    'url': ep_k,
                    'auth': ep_v['auth'],
                    'token': ep_v['token'],
                }
                ep = ConfigEndpoint.from_json(json.dumps(t_ep))
                endpoints.append(ep)
            ep = len(endpoints)
            _LOGGING.debug(
                f'Successfully loaded {ep} endpoints '
                f'from legacy configuration.'
            )
            confirmation = confirm or click.confirm(
                f'\nWould you like to upgrade {ep} endpoint(s)? '
                f'This action will \n'
                f'create a new configuration file {ctx.config} \n'
                f'with your endpoints in it'
            )
            if confirmation:
                # load new configuration file
                config_file = ctx.load_config_template()
                config_file.update_endpoints(*endpoints)
                # check if target exists
                target_exists = os.path.isfile(ctx.config)
                if target_exists:
                    if not (
                        overwrite
                        or click.confirm(f'\nOverwrite {ctx.config}?')
                    ):
                        raise click.Abort('Cancelled by user')
                # all ok
                ctx.write_config_file(new_config_file=config_file)
                ctx.secho(
                    f'\nSuccessfully migrated {legacy_config} {ej_tada}',
                    fg='green',
                    nl=True,
                )
            else:
                raise click.Abort('Cancelled by user')
        else:
            _LOGGING.warning('No endpoints found in configuration file.')
            cfg_exc = True
    except KeyError as ex:
        _LOGGING.warning(f'Missing {str(ex)} in endpoint configuration.')
        cfg_exc = True
    except (TypeError, ParserError) as ex:
        _LOGGING.warning(f'{str(ex)}')
        cfg_exc = True
    finally:
        if cfg_exc:
            _LOGGING.warning(
                'Try running "vss-cli configure mk" '
                'to create a new configuration file.'
            )


@cli.command('mk', short_help='Create new endpoint configuration.')
@click.option(
    '-r',
    '--replace',
    is_flag=True,
    default=False,
    help='Replace existing configuration',
)
@click.option(
    '-e',
    '--endpoint-name',
    type=click.STRING,
    help='Custom endpoint name. Default to endpoint hostname.',
    required=False,
)
@pass_context
def mk(ctx: Configuration, replace: bool, endpoint_name: str):
    """Create new configuration or add profile to config file"""
    new_endpoint = ctx.endpoint or click.prompt(
        'Endpoint',
        default=const.DEFAULT_ENDPOINT,
        type=click.STRING,
        show_default=True,
    )
    endpoint_name = endpoint_name or click.prompt(
        'Endpoint Name',
        default=ctx.endpoint_name,
        type=click.STRING,
        show_default=True,
    )
    username = ctx.username or click.prompt(
        'Username', default=ctx.username, show_default=True, type=click.STRING
    )
    password = ctx.password or click.prompt(
        'Password',
        default=ctx.password,
        show_default=False,
        hide_input=True,
        type=click.STRING,
        confirmation_prompt=True,
    )
    is_configured = ctx.configure(
        username=username,
        password=password,
        endpoint=new_endpoint,
        replace=replace,
        endpoint_name=endpoint_name,
    )
    if is_configured:
        # feedback message
        ctx.secho(f'You are ready to use the vss-cli {ej_rkt}', fg='green')
    else:
        _LOGGING.warning(
            f'Houston, we have a problem {ej_warn}. '
            f'Could not create configuration.'
        )


COLUMNS_DETAILS = [
    ("NAME", "name"),
    ("ENDPOINT", "endpoint"),
    ("USER", "user"),
    ("PASS", "pass"),
    ("TOKEN", "token"),
    ("SOURCE", "source"),
    ('DEFAULT', 'default'),
]


@cli.command('set', short_help='Update general settings attribute.')
@click.argument('setting', type=click.Choice(const.GENERAL_SETTINGS.keys()))
@click.argument('value', type=click.STRING)
@pass_context
def set_cfg(ctx: Configuration, setting: str, value: Any):
    ctx.load_config(validate=False)
    data_type = const.GENERAL_SETTINGS[setting]
    was = None
    to = None
    try:
        was = getattr(ctx.config_file.general, setting)
        f = str2bool if data_type == bool else data_type
        to = f(value)
        if setting == 'default_endpoint_name':
            is_ok = ctx.config_file.get_endpoint(value)
            if not is_ok:
                raise click.BadArgumentUsage(
                    f'Endpoint {value} does not exist. '
                    f'Run "vss-cli configure mk -e {value}" '
                    f'to create endpoint first.'
                )
        setattr(ctx.config_file.general, setting, to)
    except ValueError:
        _LOGGING.warning(f'{setting} value must be {data_type}')
    ctx.secho(f"Updating {setting} from {was} -> {to}.")
    ctx.write_config_file(config_general=ctx.config_file.general)
    ctx.secho(f"{ctx.config} updated {ej_save}", fg='green')
    return


@cli.command('ls', short_help='List existing endpoint configuration')
@pass_context
def ls(ctx: Configuration):
    """List existing configuration"""
    from base64 import b64decode

    cfg_endpoints = list()
    try:
        config_file = ctx.load_config_file()
        ctx.set_defaults()
        default_endpoint = config_file.general.default_endpoint_name
        endpoints = config_file.endpoints or []
        # checking profiles
        for endpoint in endpoints:
            is_default = ej_check if default_endpoint == endpoint.name else ''
            token = ''
            user = ''
            pwd = ''
            auth = endpoint.auth
            url = endpoint.url
            if auth:
                auth_enc = auth.encode()
                user, pwd = b64decode(auth_enc).split(b':')
                user = user.decode()
            masked_pwd = ''.join(['*' for i in range(len(pwd))])
            if endpoint.token:
                token = f"{endpoint.token[:10]}...{endpoint.token[-10:]}"

            cfg_endpoints.append(
                {
                    'default': is_default,
                    'name': endpoint.name,
                    'endpoint': url,
                    'user': user,
                    'pass': masked_pwd[:8],
                    'token': token,
                    'source': 'config file',
                }
            )
    except FileNotFoundError as ex:
        _LOGGING.error(f'{str(ex)}')
        ctx.secho(f'Have you run ' f'"vss-cli configure mk"?', fg='green')

    # checking env vars
    envs = [e for e in os.environ if 'VSS_' in e]
    if envs:
        user = os.environ.get('VSS_USER', '')
        pwd = os.environ.get('VSS_USER_PASS', '')
        masked_pwd = ''.join(['*' for i in range(len(pwd))])
        tk = os.environ.get('VSS_TOKEN', '')
        endpoint = os.environ.get('VSS_ENDPOINT', const.DEFAULT_ENDPOINT)
        source = 'env'
        cfg_endpoints.append(
            {
                'endpoint': endpoint,
                'user': user,
                'pass': masked_pwd,
                'token': '{}...{}'.format(tk[:10], tk[-10:]),
                'source': source,
            }
        )
    if cfg_endpoints:
        click.echo(format_output(ctx, cfg_endpoints, columns=COLUMNS_DETAILS))
    else:
        ctx.echo('No configuration was found')


@cli.command('edit', short_help='Edit configuration file.')
@click.option(
    '-l', '--launch', is_flag=True, help='Open config file with default editor'
)
@pass_context
def edit(ctx: Configuration, launch):
    """Edit configuration file"""
    # either launch or edit in
    if launch:
        click.launch(ctx.config, locate=True)
    else:
        # proceed to load file
        with open(ctx.config, 'r') as data_file:
            raw = data_file.read()
        # launch editor
        new_raw = click.edit(raw, extension='.yaml')
        if new_raw is not None:
            ctx.secho(f"Updating {ctx.config} {ej_save}", fg='green')
            new_obj = ctx.yaml_load(new_raw)
            with open(ctx.config, 'w') as fp:
                ctx.yaml_dump_stream(new_obj, stream=fp)
        else:
            ctx.echo("No edits/changes returned from editor.")
            return



"""VSS Service Management plugin for VSS CLI (vss-cli)."""
import click

from vss_cli import const, rel_opts as so
from vss_cli.cli import pass_context
from vss_cli.config import Configuration
from vss_cli.helper import format_output


@click.group('service', short_help='ITS Service catalog.')
@pass_context
def cli(ctx: Configuration):
    """Available ITS Service catalog."""
    with ctx.spinner(disable=ctx.debug):
        ctx.load_config()


@cli.command('ls', short_help='list available ITS Service catalog.')
@so.filter_opt
@so.sort_opt
@so.all_opt
@so.count_opt
@so.page_opt
@pass_context
def service_ls(ctx: Configuration, filter_by, page, sort, show_all, count):
    """List services based on:

        Filter list in the following format <field_name> <operator>,<value>
        where operator is eq, ne, lt, le, gt, ge, like, in.
        For example: name=like,%VPN%

            vss-cli service ls -f name=like,%VPN%

        Sort list in the following format <field_name>=<asc|desc>. For example:

            vss-cli service ls -s label=desc

    """
    columns = ctx.columns or const.COLUMNS_VSS_SERVICE
    params = dict()
    if all(filter_by):
        params['filter'] = ';'.join(filter_by)
    if all(sort):
        params['sort'] = ';'.join(sort)
    # make request
    with ctx.spinner(disable=ctx.debug):
        _requests = ctx.get_vss_services(
            show_all=show_all, per_page=count, **params
        )
    # format output
    output = format_output(ctx, _requests, columns=columns)
    # page results
    if page:
        click.echo_via_pager(output)
    else:
        click.echo(output)
